/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef _UDM_DB_INT_H
#define _UDM_DB_INT_H

#include <limits.h>
#include "udm_sqldbms.h"
#include "udm_indexcache.h"
#include "udm_searchtool.h" /* UDM_QUERY_PARAM */

/* Multi-dict mode defines */
#define MULTI_DICTS	0xff

#define	URL_DELETE_CACHE		256
#define	URL_SELECT_CACHE		256
#define URL_LOCK_TIME			4*60*60
#define UDM_MAX_MULTI_INSERT_QSIZE	16*1024
#define NDOCS_QUERY			"SELECT count(*) FROM url"


/********************************************/

typedef struct
{
  char *word;
  UDM_URLID_COORD urlid_coord;
  unsigned char seed;
} UDM_WORD_CACHE_WORD;


typedef struct
{
  size_t nbytes;
  size_t nwords;
  size_t awords;
  UDM_WORD_CACHE_WORD *words;
  size_t nurls;
  size_t aurls;
  urlid_t *urls;
} UDM_WORD_CACHE;

typedef struct
{
  char    *url[UDM_FINDURL_CACHE_SIZE];
  urlid_t id[UDM_FINDURL_CACHE_SIZE];
  size_t  pURLCache;
} UDM_URLID_CACHE;


/*******************************************/

/* SQL storage modes */
typedef enum
{
  UDM_SQLDBMODE_SINGLE=  0,
  UDM_SQLDBMODE_MULTI=   1,
  UDM_SQLDBMODE_BLOB=    2,
  UDM_SQLDBMODE_RAWBLOB= 3
} udm_sqldbmode_t;


struct udm_dbmode_handler_st;

typedef struct udm_sqldb_st
{
  UDM_SQL sql;
  const struct udm_dbmode_handler_st *dbmode_handler;
  udm_sqldbmode_t DBMode;
  time_t last_notargets_time;
  UDM_WORD_CACHE WordCache;
  UDM_URLID_CACHE URLIdCache;
} UDM_SQLDB;


udm_sqldbtype_t UdmSQLDBType(UDM_DB *db);
udm_sqldbapi_t UdmSQLDBDriver(UDM_DB *db);
int UdmSQLDBFlags(UDM_DB *db);
udm_sqldbmode_t UdmSQLDBMode(UDM_DB *db);
int UdmSQLDBVersion(UDM_DB *db);
UDM_SQLDB_HANDLER *UdmSQLDBHandler(UDM_DB *db);
const char *UdmSQLDBQueryFrom(UDM_QUERY *query);
UDM_SQLDB *UdmSQLDB(UDM_DB *db);
int UdmSQLDBConnected(UDM_DB *db);
UDM_VARLIST *UdmSQLDBVars(UDM_DB *db);

udm_rc_t UdmDBSQLQuery(UDM_AGENT *A, UDM_DB *db, UDM_SQLRES *Res, const char *buf);
udm_rc_t UdmDBSQLDropTableIfExists(UDM_AGENT *A, UDM_DB *db, const char *name);
udm_rc_t UdmDBSQLTableTruncateOrDelete(UDM_AGENT *A, UDM_DB *db, const char *name);
udm_rc_t UdmDBSQLCopyStructure(UDM_AGENT *A, UDM_DB *db, const char *from, const char *to);
udm_rc_t UdmDBSQLRenameTable(UDM_AGENT *A, UDM_DB *db, const char *from, const char *to);
udm_rc_t UdmDBSQLBegin(UDM_AGENT *A, UDM_DB *db);
udm_rc_t UdmDBSQLCommit(UDM_AGENT *A, UDM_DB *db);
udm_rc_t UdmDBSQLExecDirect(UDM_AGENT *A, UDM_DB *db, UDM_SQLRES *R, const char *query);
udm_rc_t UdmDBSQLPrepare(UDM_AGENT *A, UDM_DB *db, const char *query);
udm_rc_t UdmDBSQLExecute(UDM_AGENT *A, UDM_DB *db);
udm_rc_t UdmDBSQLFetchRow(UDM_AGENT *A, UDM_DB *db, UDM_SQLRES *R, UDM_STR *);
udm_rc_t UdmDBSQLStmtFree(UDM_AGENT *A, UDM_DB *db);
udm_rc_t UdmDBSQLBindParameter(UDM_AGENT *A, UDM_DB *db, int pos, const void *data, int size, udm_sqltype_t type);
udm_rc_t UdmDBSQLLockOrBegin(UDM_AGENT *A, UDM_DB *db, const char *param);
udm_rc_t UdmDBSQLUnlockOrCommit(UDM_AGENT *A, UDM_DB *db);
char*  UdmDBSQLEscStrAlloc(UDM_AGENT *A, UDM_DB *db, const char *src, size_t srclen);
char*  UdmDBSQLEscStrSimple(UDM_AGENT *A, UDM_DB *db, char *to, const char *from, size_t l);
size_t UdmDBSQLBinEscStr(UDM_AGENT *A, UDM_DB *db, char *dst, size_t dstlen, const char *src, size_t srclen);
udm_rc_t UdmDBSQLQueryOneRowInt(UDM_AGENT *A, UDM_DB *db, int *res, const char *qbuf);
size_t UdmDBSQLEscStr(UDM_AGENT *A, UDM_DB *db, char *to, const char *from, size_t l);
void UdmDBSQLTopClause(UDM_AGENT *A, UDM_DB *db, size_t top_num, UDM_SQL_TOP_CLAUSE *Top);

const char *UdmDBSQLParamPlaceHolder(UDM_DB *db, size_t i);
char *UdmDBSQLError(UDM_DB *db);
size_t UdmDBSQLErrorSize(UDM_DB *db);

/******************/

typedef struct st_udm_findword_args
{
  UDM_URLID_LIST urls; /* Combination of WHERE limit, fl, and live_deleted */
  UDM_URLID_LIST live_update_active_urls;  /* Combination of WHERE and fl */
  UDM_URLID_LIST live_update_deleted_urls; /* The list of deleted */
  UDM_WIDEWORDLIST CollationMatches;
  UDM_SEARCHSECTIONLISTLIST SearchSectionListList;
  const char *cmparg;
  const char *where;
  UDM_WIDEWORD Word;
  int live_updates;
  int need_coords; /* 0 for regular word, 1 for multi-word */
  time_t live_updates_ts;
  UDM_SQLRESLIST SQLResults;
  UDM_INVERTED_INDEX_CACHE IndexCache;
  UDM_QUERY_PARAM query_param;
} UDM_FINDWORD_ARGS;


typedef struct udm_dbmode_handler_st
{
  const char *name;
  udm_rc_t (*StoreWords)(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc);
  udm_rc_t (*QueryAction)(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query, udm_querycmd_t cmd);
  udm_rc_t (*DeleteWordsFromURL)(UDM_AGENT *Indexer, UDM_DB *db, urlid_t url_id);
  udm_rc_t (*FindWord)(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query, UDM_FINDWORD_ARGS *args);
  udm_rc_t (*DumpWordInfo)(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc);
  udm_rc_t (*InitSearch)(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query, UDM_FINDWORD_ARGS *args);
} UDM_DBMODE_HANDLER;

extern const UDM_DBMODE_HANDLER udm_dbmode_handler_single;
extern const UDM_DBMODE_HANDLER udm_dbmode_handler_multi;
extern const UDM_DBMODE_HANDLER udm_dbmode_handler_blob;
extern const UDM_DBMODE_HANDLER udm_dbmode_handler_rawblob;

/* Functions from sql.c */
udm_rc_t UdmLoadSlowLimit(UDM_AGENT *A, UDM_DB *db, UDM_URLID_LIST *list, const char *q);
udm_rc_t UdmUserScoreListLoad(UDM_AGENT *A, UDM_DB *db,
                              UDM_URL_INT4_LIST *List, const char *q);

/* Functions from dbmode-blob.c */
udm_rc_t UdmBlobLoadFastURLLimit(UDM_AGENT *A, UDM_DB *db, const char *name,
                                 UDM_URLID_LIST *buf);
udm_rc_t UdmBlobLoadFastOrder(UDM_AGENT *A, UDM_DB *db, UDM_URL_INT4_LIST *,const char *name);
udm_rc_t UdmBlobLoadFastScore(UDM_AGENT *A, UDM_DB *db, UDM_URL_INT4_LIST *,const char *name);
udm_rc_t UdmLoadURLDataFromBdict(UDM_AGENT *A, UDM_DB *db,
                                 UDM_URLDATALIST *DataList, int flags);
udm_rc_t UdmBlobReadTimestamp(UDM_AGENT *A, UDM_DB *db, int *ts, int def);

/* Functions from dbmode-rawblob.c */
#define UDM_RAWBLOB_SEARCH  0
#define UDM_RAWBLOB_DELTA   1
udm_rc_t UdmFindWordRawBlobDelta(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                                 UDM_FINDWORD_ARGS *args);

/* Functions from wordcache.c */
udm_rc_t UdmWordCacheWrite(UDM_AGENT *Indexer, UDM_DB *db, size_t limit);
udm_rc_t UdmWordCacheAddWordList(UDM_WORD_CACHE *Cache,
                                 UDM_WORDLIST *List, urlid_t id);
UDM_WORD_CACHE *UdmWordCacheInit (UDM_WORD_CACHE *cache);
void UdmWordCacheFree(UDM_WORD_CACHE *cache);
void UdmWordCacheSort(UDM_WORD_CACHE *cache);
udm_rc_t UdmWordCacheAdd(UDM_WORD_CACHE *cache, urlid_t url_id, UDM_WORD *W);
udm_rc_t UdmWordCacheAddURL (UDM_WORD_CACHE *cache, urlid_t url_id);

/* Functions from suggest.c */
udm_rc_t UdmWordStatQuery(UDM_AGENT *A, UDM_DB *db, const char *src);
udm_rc_t UdmQuerySuggest(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query);

/* Functions from qcache.c */
udm_rc_t UdmQueryCachePutSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query);
udm_rc_t UdmQueryCacheGetSQL(UDM_AGENT *Indexer, UDM_DB *db, UDM_QUERY *Query);
udm_rc_t UdmApplyCachedQueryLimit(UDM_AGENT *query,
                                  UDM_URLSCORELIST *ScoreList, UDM_DB *db);

#endif
