/* CRC32 checksum */

#include <stdio.h>
#include <sys/types.h>

#include "udm_crc32.h"

#include "crc32int.h"


udmcrc32_t UdmCRC32(const char *b, size_t size)
{
  udmcrc32_t crc= (udmcrc32_t)~0L;
  const char *e;
  
  for (e= b + size; b < e; ++b)
    UDM_CRC32_UPDATE(crc, *b) ;
  return (~crc);
}

udmcrc32_t UdmCRC32Update(int prev,const char *b, size_t size)
{
  udmcrc32_t crc= (udmcrc32_t)~prev;
  const char *e;
  
  for (e= b + size; b < e; ++b)
    UDM_CRC32_UPDATE(crc, *b);
  return (~crc);
}


/* Update previous crc32 value with a null-terminated string */
udmcrc32_t
UdmCRC32UpdateStr(udmcrc32_t prev, const char *s)
{
  prev= ~prev;
  for ( ; *s; s++)
    UDM_CRC32_UPDATE(prev, *s);
  return (~prev);
}


udmcrc32_t UdmCRC32UpdateUnicode(int prev, const int *b, size_t charlen)
{
  udmcrc32_t crc= (udmcrc32_t)~prev;
  const int *e;
  
  for (e= b + charlen; b < e; ++b)
  {
    char p1, p2;
    p1= (char) (*b >> 8) & 0xFF;
    p2= (char) (*b & 0xFF);
    UDM_CRC32_UPDATE(crc, p1);
    UDM_CRC32_UPDATE(crc, p2);
  }
  return (~crc);
}




#ifdef UDM_CRC32_TEST
int main()
{
  char c[]="12345678";
  char a[]="1234";
  char b[]="5678";
  char d[256];
  int res;
  
  for (res=0; res < 256; (unsigned char)(d[res])= res++);
  
  /* Direct */
  res=UdmCRC32(c,strlen(c));
  printf("%08X\n",res);

  /* Incremental */
  res=0;
  res=UdmCRC32Update(res,a,strlen(a));
  res=UdmCRC32Update(res,b,strlen(b));
  printf("%08X\n",res);
  
  /* A longer array */
  res=UdmCRC32(d, 256);
  printf("%08X\n",(int) res);
}
#endif
