/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#ifdef WIN32
#include <io.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_execget.h"
#include "udm_log.h"
#include "udm_http.h"

/*
#define DEBUG_EXECGET
*/

udm_rc_t
UdmExecGet(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc)
{
  char * args;
  char cmdline[1024];
  FILE * f;
  UDM_URL *Url=&Doc->CurURL;

  UdmHTTPBufReset(&Doc->Buf);

  if((args = strchr(UDM_NULL2EMPTY(Url->filename), '?')))
  {
    *args='\0';
    args++;
  }
  
  if (!strcmp(UDM_NULL2EMPTY(Url->schema), "exec") && args)
    udm_snprintf(cmdline, sizeof(cmdline), "%s%s \"%s\"",
                 UDM_NULL2EMPTY(Url->path), UDM_NULL2EMPTY(Url->filename), args);
  else
    udm_snprintf(cmdline, sizeof(cmdline), "%s%s",
                 UDM_NULL2EMPTY(Url->path), UDM_NULL2EMPTY(Url->filename));

  if(!strcmp(UDM_NULL2EMPTY(Url->schema), "cgi"))
  {
    /* Non-parsed-headers CGI return HTTP status itself */
    if(strncasecmp(UDM_NULL2EMPTY(Url->filename), "nph-", 4))
    {
      UdmHTTPBufAppendf(&Doc->Buf, "HTTP/1.0 200 OK\r\n");
    }
    UdmSetEnv("QUERY_STRING",args?args:"");
    UdmSetEnv("REQUEST_METHOD","GET");
  }

#ifdef DEBUG_EXECGET
  fprintf(stderr,"cmd='%s'\n",cmdline);
#endif

  UdmLog(Indexer,UDM_LOG_DEBUG,"Starting program '%s'",cmdline);

  f= popen(cmdline,"r");

  /* Remove set variables */
  if(!strcmp(UDM_NULL2EMPTY(Url->schema), "cgi"))
  {
    UdmUnsetEnv("REQUEST_METHOD");
    UdmUnsetEnv("QUERY_STRING");
  }

  if (f)
  {
    UdmHTTPBufAppendFromFile(&Doc->Buf, fileno(f));
    pclose(f);
  }
  else
  {
    int status;
    UdmLog(Indexer, UDM_LOG_ERROR, "UdmExecGet failed: error=%s", strerror(errno));
    switch(errno)
    {
      case ENOENT: status= 404; break; /* Not found */
      case EACCES: status= 403; break; /* Forbidden*/
      default: status= 500;
    }
    UdmHTTPBufReset(&Doc->Buf);
    UdmHTTPBufAppendf(&Doc->Buf, "HTTP/1.0 %d %s\r\n\r\n",
                      status, strerror(errno));
  }
#ifdef DEBUG_EXECGET
  fprintf(stderr, "buf='%s'\n", Indexer->buf);
#endif
  return UDM_OK;
}
