/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"

#include "udm_utils.h"
#include "udm_uniconv.h"
#include "udm_searchtool.h"
#include "udm_result.h"
#include "udm_vars.h"

#if 0
static UDM_WIDEWORD*
UdmWordInWWList(UDM_WIDEWORDLIST *List, int *tok, size_t toklen,
                int hlstop, size_t phrpos)
{
  size_t uw;
  
  for(uw=0; uw < List->nwords; uw++)
  {
    size_t slen;
    UDM_WIDEWORD *W= &List->Word[uw];
    if (W->phrpos != phrpos)
      continue;
    if (!hlstop && W->origin == UDM_WORD_ORIGIN_STOP)
      continue;
    slen= W->ulen;
    if (wordlen < slen)
      continue;
    if (wordlen == slen && !UdmUniStrNCaseCmp(tok, W->uword, slen))
      return W;
      
    if (wordlen > slen) switch (W->match)
    {
      case UDM_MATCH_BEGIN:
        if (!UdmUniStrNCaseCmp(tok, W->uword, slen))
          return W;
        break;
      case UDM_MATCH_END:
        if (!UdmUniStrNCaseCmp(tok + wordlen - slen, W->uword, slen))
          return W;
        break;
      case UDM_MATCH_SUBSTR:
        {
          size_t l1, l2;
          for (l1 = 0; l1 < wordlen; l1++)
          {
            if (l1 + slen > wordlen) break;
            for (l2 = 0; l2 < slen; l2++)
            {
              if (UdmUniToLower(tok[l1 + l2]) != UdmUniToLower(W->uword[l2]))
                break;
            }
            if (l2 == slen)
            {
              return W;
              break;
            }
          }
        }
        break;
    }
  }
  return NULL;
}
#endif


void
UdmExcerptConvInit(UDM_HIGHLIGHT_CONV *cnv,
                   UDM_CHARSET *wcs, UDM_CHARSET *src, UDM_CHARSET *dst,
                   int hlstop, int segmenter)
{
  UdmConvInit(&cnv->uni_wcs, &udm_charset_sys_int, wcs);
  UdmConvInit(&cnv->src_uni, src, &udm_charset_sys_int);
  UdmConvInit(&cnv->uni_dst, &udm_charset_sys_int, dst);
  cnv->hlstop= hlstop;
  cnv->segmenter= segmenter;
}


void
UdmExcerptConvInitFromEnv(UDM_HIGHLIGHT_CONV *cnv,
                          UDM_CHARSET *wcs, UDM_CHARSET *src, UDM_CHARSET *dst,
                          UDM_ENV *Env)
{
  int hlstop= UdmVarListFindBool(&Env->Vars, "ExcerptStopword", UDM_TRUE);
  const char *seg= UdmVarListFindStr(&Env->Vars, "Segmenter", NULL);
  int segmenter= seg ? UdmUniSegmenterFind(Env, NULL, seg) : 0;
  UdmExcerptConvInit(cnv, wcs, src, dst, hlstop, segmenter);
}

static UDM_WIDEWORD*
UdmWordInWWList2(UDM_WIDEWORDLIST *List,
                 int *tok, size_t toklen,
                 UDM_CONV *uni_wcs, UDM_CONV *lc_uni, UDM_CONV *uni_bc,
                 int hlstop, size_t phrpos)
{
  size_t uw;
  char word[128]= "";
  size_t wordlen= 0;
  UDM_UNIDATA *unidata= udm_unidata_default; /* FIXME */

  if (List->match_mode != UDM_MATCH_FULL)
  {
    wordlen= UdmStrToLowerExt(unidata, uni_wcs, word, sizeof(word) - 1,
                              (char *) tok, toklen * sizeof(int),
                              UDM_RECODE_HTML);
    word[wordlen]= '\0';
  }
  
  for(uw=0; uw < List->nwords; uw++)
  {
    size_t W_len;
    UDM_WIDEWORD *W= &List->Word[uw];

    if (W->Param.phrpos != phrpos)
      continue;
    if (!hlstop && W->Param.origin == UDM_WORD_ORIGIN_STOP)
      continue;

    if (W->Param.match_mode == UDM_MATCH_FULL)
    {
      int cmp= List->strip_noaccents ?
          UdmStrCaseAccentCmp2(unidata, uni_wcs,
                               (const char*) tok, toklen * sizeof(int),
                               W->Word.str, W->Word.length, UDM_RECODE_HTML) :
          UdmStrCaseCmp2(unidata, uni_wcs,
                         (const char*) tok, toklen*sizeof(int),
                         W->Word.str, W->Word.length, UDM_RECODE_HTML);
      if (!cmp)
        return W;
      continue;
    }

    W_len= W->Word.length;

    if (wordlen < W_len)
      continue;
    if (wordlen == W_len && !memcmp(word, W->Word.str, W_len))
      return W;

    if (wordlen > W_len) switch (W->Param.match_mode)
    {
      case UDM_MATCH_BEGIN:
        if (!memcmp(word, W->Word.str, W_len))
          return W;
        break;
      case UDM_MATCH_END:
        if (!memcmp(word + wordlen - W_len, W->Word.str, W_len))
          return W;
        break;
      case UDM_MATCH_SUBSTR:
        if (strstr(word, W->Word.str))
          return W;
        break;
      case UDM_MATCH_FULL:
      case UDM_MATCH_REGEX:
      case UDM_MATCH_WILD:
      case UDM_MATCH_SUBNET:
      case UDM_MATCH_NUMERIC_LT:
      case UDM_MATCH_NUMERIC_GT:
      case UDM_MATCH_RANGE:
        UDM_ASSERT(0); /* Inpossible */
    }
  }
  return NULL;
}


/*
  Remove hilight markers from a string and return
  its new length, in bytes.
*/
size_t
UdmRemoveHl(UDM_CHARSET *cs, char *str, size_t from, size_t to)
{
  if (cs == &udm_charset_sys_int)
  {
    int *stri= (int*) str;
    int *s= stri + from / sizeof(int);
    int *e= stri + to / sizeof(int);
    int *dst= s;
    for (; s < e; s++)
    { 
      if (*s != 2 && *s != 3)
        *dst++= *s;
    }
    return (dst - stri) * sizeof(int);
  }
  else
  {
    char *s= str + from, *e= str + to, *dst= s;
    for (; s < e; s++)
    { 
      if (*s != 2 && *s != 3)
        *dst++= *s;
    }
    return dst - str;
  }
}


static void
UdmHlAppend(UDM_CONV *uni_bc, UDM_WIDEWORD *found,
            UDM_DSTR *dstr, const int *tok, size_t toklen)
{
  int i2= 2, i3= 3;

  if (found)
    UdmDSTRAppendConv(dstr, uni_bc, (char*) &i2, sizeof(i2), UDM_RECODE_HTML);
  if (uni_bc->to == &udm_charset_sys_int)
    UdmDSTRAppend(dstr, (const char*) tok, sizeof(*tok) * toklen);
  else
    UdmDSTRAppendConv(dstr, uni_bc, (const char*) tok, sizeof(*tok) * toklen, UDM_RECODE_HTML);
  if (found)
    UdmDSTRAppendConv(dstr, uni_bc, (char*) &i3, sizeof(i3), UDM_RECODE_HTML);
  /*fprintf(stderr, "appended to '%.*s'\n", dstlen, dst);*/
}


void UdmSnippetChunkStatInit(UDM_SNIPPETCHUNK_STAT *st)
{
  st->mask= 0;
  st->quality= 0;
}


static void UdmSnippetChunkStatAdd(UDM_SNIPPETCHUNK_STAT *st,
                                   UDM_WIDEWORD *found)
{
  udm_found_word_bitmask_t curmask= 1ULL << found->Param.order;
  if (!(st->mask & curmask))
  {
    st->mask|= curmask;
    st->quality+= found->Param.weight;
  }
}


/*
#define DEBUG_HL 0
*/

/*
  Convert "src" to "dstr", with searched word highlighting.
*/

size_t
UdmHlConvertExtWithConv(UDM_AGENT *Agent,
                        UDM_DSTR *dstr,
                        UDM_SNIPPETCHUNK_STAT *found_stat,
                        UDM_WIDEWORDLIST *List,
                        const char *src, size_t srclen,
                        UDM_HIGHLIGHT_CONV *ec)
{
  int		*tok, *lt, ctype, *uni, *uend;
  size_t        dstlen_phr= UdmDSTRLength(dstr), nfound=0, ulen;
  size_t        unimaxlen, expected_phrpos= 0;
  UDM_UNIDATA *unidata= udm_unidata_default; /* FIXME */
  UDM_SNIPPETCHUNK_STAT found_stat_prev;

#ifdef DEBUG_HL
  fprintf(stderr, "wcs=%s fromcs=%s tocs=%s srclen=%d src='%s'\n",
          uni_wcs->to->name, lc_uni->from->name, uni_bc->to->name, srclen, src);
#endif

  UdmSnippetChunkStatInit(&found_stat_prev);
  UdmSnippetChunkStatInit(found_stat);

  /* Convert to unicode */
  unimaxlen= (srclen + 10) * sizeof(int);
  uni= (int *)UdmMalloc(unimaxlen);
  ulen= UdmConvHTML(&ec->src_uni,(char*)uni, unimaxlen, src, srclen) / sizeof(int);
  uni[ulen]= '\0';
  if (ec->segmenter)
  {
    uni= UdmUniSegmentByType(Agent, uni, ec->segmenter, 0x09);
    ulen= UdmUniLen(uni);
  }
  
  uend= uni + ulen;

  /* Parse unicode string */
  for (tok= UdmUniGetSepToken(unidata, uni, uend, &lt, &ctype) ; tok ;
       tok= UdmUniGetSepToken(unidata, NULL, uend, &lt, &ctype))
  {
    size_t toklen= lt - tok;

    if (ctype == UDM_UNI_SEPAR || !List)
    {
      /* Don't append delimiters added by segmenter */
      if (ec->segmenter && toklen == 1 && *tok == 0x09)
        continue;
      UdmHlAppend(&ec->uni_dst, NULL, dstr, tok, toklen);
    }
    else
    {
      UDM_WIDEWORD *found= UdmWordInWWList2(List, tok, toklen,
                                            &ec->uni_wcs, &ec->src_uni, &ec->uni_dst,
                                            ec->hlstop, expected_phrpos);
      if (found)
      {
        UdmSnippetChunkStatAdd(found_stat, found);
        UdmHlAppend(&ec->uni_dst, found, dstr, tok, toklen);
        nfound++;
        if (found->Param.phrpos + 1 == found->Param.phrlen)
        {
          /* last in phrase found */
          expected_phrpos= 0;
          dstlen_phr= UdmDSTRLength(dstr);
          found_stat_prev= *found_stat;
          nfound= 0;
        }
        else
        {
          /* middle in phrase found */
          expected_phrpos++;
        }
      }
      else
      {
        /* No word found on expected phrase position, rollback */
        if (nfound)
        {
          *found_stat= found_stat_prev;
          UdmDSTRRemoveHighlightWithOffset(dstr, ec->uni_dst.to, dstlen_phr);
          /*
            Search the same word on the first position,
            it can start new phrase.
            For example:
            Document: "a a b b"
            Query: "a-b"
          */
          found= UdmWordInWWList2(List, tok, toklen,
                                  &ec->uni_wcs, &ec->src_uni, &ec->uni_dst,
                                  ec->hlstop, 0);
          if (found)
          {
            UdmSnippetChunkStatAdd(found_stat, found);
            dstlen_phr= UdmDSTRLength(dstr);
            expected_phrpos= 1;
            nfound= 1;
            UdmHlAppend(&ec->uni_dst, found, dstr, tok, toklen);
            continue;
          }
        }
        
        UdmHlAppend(&ec->uni_dst, NULL, dstr, tok, toklen);
        dstlen_phr= UdmDSTRLength(dstr);
        expected_phrpos= 0;
        nfound= 0;
      }
    }
  }

#ifdef DEBUG_HL  
  fprintf(stderr, "end: expected_phrpos=%d dstlen=%d dstlen_phr=%d\n", expected_phrpos, dstlen, dstlen_phr);
#endif
  if (expected_phrpos > 0)
  {
    /* Roll back: incomplete last phrase */
    UdmDSTRRemoveHighlightWithOffset(dstr, ec->uni_dst.to, dstlen_phr);
    *found_stat= found_stat_prev;
  }

  UdmFree(uni);
  return UdmDSTRLength(dstr);
}


#ifdef NOT_YET
/*
  For character sets having "septok" and when lcs == bcs
  
  TODO: this function does not support HTML entities:
    &Auml;
    &#196;
    &#xC4;

  UdmWordInWWList2() should also be modified to unserstand entitites
  when searching words.
*/
static char *
UdmHlConvertExtQuick(UDM_WIDEWORDLIST *List, const char *src,
                     UDM_CHARSET *cs,
                     int hilight_stopwords)
{
  char        *dst;
  const char  *srcend, *tok, *lt;
  size_t      srclen, dstlen= 0, dstlen_phr= 0;
  size_t      dstmaxlen, expected_phrpos= 0;
  int ctype;

  if(!src || !(srclen = strlen(src)))
    return NULL;

  srcend= src + srclen;
  dstmaxlen= srclen * 14 + 10;
  dst= (char*)UdmMalloc(dstmaxlen + 1);

  for (tok= cs->septoken(cs, src, srcend, &lt, &ctype) ; tok ;
       tok= cs->septoken(cs, NULL, srcend, &lt, &ctype))
  {
    size_t toklen= lt - tok;

    if (ctype == UDM_UNI_SEPAR || !List || toklen > 127)
    {
      memcpy(dst + dstlen, tok, toklen);
      dstlen+= toklen;
    }
    else
    {
      UDM_WIDEWORD *found;
      {
        char tmp[128];
        memcpy(tmp, tok, toklen);
        tmp[toklen]= '\0';
        cs->lcase(cs, tmp, toklen);
        found= UdmWordInWWList2(List, tmp, toklen, hilight_stopwords, expected_phrpos);
      }

      if (dstlen + toklen + 2 >= dstmaxlen)
        break;
      
      if (found)
      {
        dst[dstlen++]= '\2';
        memcpy(dst + dstlen, tok, toklen);
        dstlen+= toklen;
        dst[dstlen++]= '\3';
      }
      else
      {
        memcpy(dst + dstlen, tok, toklen);
        dstlen+= toklen;
      }
      
      if (found)
      {
        if (found->phrpos + 1 == found->phrlen)
        {
          /* last in phrase found */
          expected_phrpos= 0;
          dstlen_phr= dstlen;
        }
        else
        {
          /* middle in phrase found */
          expected_phrpos++;
        }
      }
      else
      {
        /* No word found on expected phrase position, rollback */
        dstlen= UdmRemoveHl(cs, dst, dstlen_phr, dstlen);
        dstlen_phr= dstlen;
        expected_phrpos= 0;
      }
    }
  }

  if (expected_phrpos > 0)
  {
    /* Roll back: incomplete last phrase */
    dstlen= UdmRemoveHl(cs, dst, dstlen_phr, dstlen);
  }

  dst[dstlen]= '\0';
  return dst;
}
#endif


udm_rc_t
UdmVarListHlConvert(UDM_AGENT *A,
                    UDM_WIDEWORDLIST *WWList,
                    UDM_VARLIST *Vars,
                    UDM_HIGHLIGHT_CONV *ec)
{
  size_t sec;

  for (sec= 0; sec < Vars->nvars; sec++)
  {
    UDM_VAR *Var= UdmVarListFindByIndex(Vars, sec);
    UDM_CONST_STR valuebuf, *value= UdmVarGetConstStr(Var, &valuebuf);
    /*
       A temporary fix to skip URL and CachedCopy:
       We will skip these sections for now.
       But this need a further fix in search.htm
       to distinguish two HTML formats:
       - HTML with &<>" escaped to &amp;&lt;&gt;&quot;
       - HTML with &<>" printed as is, no word hilight
       - HTML with &<>" printed as is, search word hilighted.
    */
    if (strcasecmp(UdmVarName(Var), "URL") &&
        strcasecmp(UdmVarName(Var), "Content-Type") &&
        value->str && value->length)
    {
      UDM_DSTR tmp;
      UDM_SNIPPETCHUNK_STAT dummy;
      /* skip if highlight markers already exist - cluster node */
      if (UdmVarFlags(Var) & UDM_VARFLAG_HL)
        continue;
      if (UDM_OK != UdmDSTRInit(&tmp, 64))
        continue;
      UdmHlConvertExtWithConv(A, &tmp, &dummy, WWList,
                              value->str, value->length, ec);
      UdmVarSetStrn(Var, UdmDSTRPtr(&tmp), UdmDSTRLength(&tmp)); /* TODO34: take over */
      UdmDSTRFree(&tmp);
    }
  }
  return UDM_OK;
  
}

            
udm_rc_t
UdmQueryConvert(UDM_AGENT *Agent, UDM_QUERY *Query,
                UDM_CHARSET *lcs, UDM_CHARSET *bcs)
{
  size_t i;
  UDM_CONV lc_bc;
  UDM_HIGHLIGHT_CONV ec;
  UDM_ENV *Conf= Agent->Conf;

  /* Init converters */
  UdmConvInit(&lc_bc, lcs, bcs);
  UdmExcerptConvInitFromEnv(&ec, bcs, lcs, bcs, Agent->Conf);

  /* Convert word list */
  for(i= 0; i < Query->Res.WWList.nwords; i++)
  {
    UDM_WIDEWORD *W= &Query->Res.WWList.Word[i];
    size_t len= strlen(W->Word.str);
    size_t newlen= UdmConvSizeNeeded(&lc_bc, len, UDM_RECODE_HTML);
    char *newval= (char*)UdmMalloc(newlen + 1);

    if (newval)
    {
      len= UdmConvHTML(&lc_bc, newval, newlen, W->Word.str, len);
      newval[len]= '\0';
      UDM_FREE(W->Word.str);
      W->Word.str= newval;
      W->Word.length= len;
    }
  }

  /* Convert document sections */
  for(i= 0; i < UdmResultNumRows(&Query->Res); i++)
  {
    UDM_DOCUMENT *D= &Query->Res.Doc[i];
    UdmVarListHlConvert(Agent, &Query->Res.WWList, &D->Sections, &ec);
    D->lcs= bcs;
  }

  /* Convert Env->Vars */
  /* TODO34: check why HLBeg and HLEnd were not converted in 3.3 */
  UdmVarListConvert(&Conf->Vars, &lc_bc, UDM_RECODE_HTML);
  return UDM_OK;
}


static char rm_hl_special[256]=
{
/*00*/  1,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,
/*10*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*20*/  0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,  /*  !"#$%&'()*+,-./ */
/*30*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* 0123456789:;<=>? */
/*40*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* @ABCDEFGHIJKLMNO */
/*50*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* PQRSTUVWXYZ[\]^_ */
/*60*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* `abcdefghijklmno */
/*70*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* pqrstuvwxyz{|}~  */
/*80*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*90*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*A0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*B0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*C0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*D0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*E0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*F0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};


size_t
UdmRemoveHiLight(char *dst, size_t dstlength,
                 const char *src, size_t srclength)
{
  size_t minlen= UDM_MIN(dstlength, srclength);
  char *dst0= dst;
  for ( ; minlen ; src++, minlen--)
  {
    if ((((unsigned char) *src) <= '&') &&
        rm_hl_special[(unsigned char) *src])
    {
      switch (src[0])
      {
        case '\0':
          goto ex;
        case '\2':
        case '\3':
          break;
        case '&':
          if (src[1] == '#')
          {
            const char *e;
            int code= 0;
          
            for (e= src + 2;
                 (*e >= '0') && (*e <= '9');
                 code= code*10 + e[0]-'0', e++)
            { }
            if (*e == ';')
            {
              *dst++= (code < 128) ? code : '?';
              src= e;
              break;
            }
          }
          /* pass through */
        
        default:
          *dst++=*src;
      }
    }
    else
      *dst++= *src;
  }
ex:
  return dst - dst0;
}


char* UdmRemoveHiLightDup(const char *s)
{
  size_t dstlength, length= strlen(s);
  char *res= (char*) UdmMalloc(length + 1);
  dstlength= UdmRemoveHiLight(res, length, s, length);
  res[dstlength]= '\0';
  return res;
}
