/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/


#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_doc.h"
#include "udm_db.h"
#include "udm_env.h"
#include "udm_agent.h"
#include "udm_store.h"
#include "udm_prog.h"
#include "udm_proto.h"
#include "udm_http.h"


typedef struct
{
  UDM_VALUE generic;
  UDM_DOCUMENT Document;
} UDM_VALUE_DOCUMENT;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_DOCUMENT Value;
} UDM_VAR_DOCUMENT;


static udm_rc_t
UdmValueDocumentDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueDocumentFlags(const char *Data)
{
  return UDM_VARFLAG_READONLY;
}


static udm_secno_t
UdmValueDocumentSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueDocumentMemUsed(const char *Data)
{
  return sizeof(UDM_DOCUMENT);
}


static void
UdmValueDocumentDestructor(char *Data)
{
  UDM_DOCUMENT *Res= (UDM_DOCUMENT *) Data;
  UdmDocFree(Res);
}


static udm_rc_t
UdmValueDocumentCopy(char *D, const char *S)
{
  UDM_ASSERT(0); /* TODO34: Add CloneVar(), with reference */
  memmove(D, S, sizeof(UDM_DOCUMENT));
  return UDM_OK;
}


static udm_rc_t
UdmValueDocumentConstructor(char *Data,
                          const UDM_VALUE **args, size_t nargs)
{
  UdmDocInit((UDM_DOCUMENT *) Data);
  return UDM_OK;
}


static size_t
UdmValueDocumentVarSize(void)
{
  return sizeof(UDM_VAR_DOCUMENT);
}


static size_t
UdmValueDocumentValueSize(void)
{
  return sizeof(UDM_VALUE_DOCUMENT);
}


static size_t
UdmValueDocumentDataOffset(void)
{
  return offsetof(UDM_VALUE_DOCUMENT, Document);
}


static size_t
UdmValueDocumentDataSize(void)
{
  return sizeof(UDM_DOCUMENT);
}


static size_t
UdmValueDocumentDataAlignment(void)
{
  return sizeof(size_t);
}


static void
UdmValueDocumentGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  static const UDM_CONST_STR name= {UDM_CSTR_WITH_LEN("#Document#")};
  *Val= name;
}


static void
UdmValueDocumentGetBool(const char *Data, udm_bool_t *val)
{
  *val= UDM_TRUE;
}


static void
UdmValueDocumentGetInt(const char *Data, int *val)
{
  *val= 0;
}


static void
UdmValueDocumentGetUInt(const char *Data, unsigned int *val)
{
  *val= 0;
}


static void
UdmValueDocumentGetDouble(const char *Data, double *val)
{
  *val= 0.0;
}


static udm_rc_t
UdmValueDocumentSetInt(char *Data, int value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDocumentSetDouble(char *Data, double value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDocumentConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDocumentAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDocumentSetStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDocumentSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                      const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDocumentAppendConv(char *Data, size_t maxlen,
                         UDM_CONV *cnv, int cnvflags,
                         const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDocumentPrintToFile(const char *Data, FILE *f)
{
  return UDM_OK;
}


static udm_rc_t
UdmValueDocumentPrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  return UDM_OK;
}


static size_t
UdmDocumentPropertyAppendHilight(UDM_DSTR *dstr,
                                 const UDM_CONST_STR *str,
                                 const char *beg, size_t blen,
                                 const char *end, size_t elen)
{
  UdmDSTRAppendHighlight(dstr, str->str, str->length, beg, blen, end, elen);
  return UDM_OK;
}


static udm_rc_t
UdmDocumentPropertyAppendHilightHtml(UDM_DSTR *dst,
                                     const UDM_CONST_STR *str,
                                     const char *hlbeg, size_t hlbeglen,
                                     const char *hlend, size_t hlendlen)
{
  UDM_DSTR tmp;
  UDM_CONV cnv;
  UdmConvInit(&cnv, &udm_charset_utf8, &udm_charset_utf8);
  UdmDSTRInit(&tmp, 128);
  UdmDSTRAppendConv(&tmp, &cnv, str->str, str->length, UDM_RECODE_HTML);
  UdmDSTRAppendHighlight(dst, UdmDSTRPtr(&tmp), UdmDSTRLength(&tmp),
                         hlbeg, hlbeglen, hlend, hlendlen);
  UdmDSTRFree(&tmp);
  return UDM_OK;
}


static UDM_VALUE *
UdmVarListFindValueByDSTR(UDM_VARLIST *Vars, const UDM_DSTR *str)
{
  char name[128];
  udm_snprintf(name, sizeof(name), "%.*s",
               (int) UdmDSTRLength(str), UdmDSTRPtr(str));
  return UdmVarListFindValue(Vars, name);
}


static void
UdmDocumentMethodPropertyCommon(UDM_PROG_EXECUTOR_STATE *state,
                                UDM_SECTION *ReturnValue,
                                const UDM_CONST_STR *str,
                                udm_bool_t html, udm_bool_t for_cached,
                                const char *hlbeg, size_t hlbeglen,
                                const char *hlend, size_t hlendlen)
{
  state->fatal_error=
    for_cached ?
    UdmDocumentPropertyAppendHilight(&ReturnValue->Value, str,
                                         hlbeg, hlbeglen,
                                         hlend, hlendlen) :  
    html ?
    UdmDocumentPropertyAppendHilightHtml(&ReturnValue->Value, str,
                                         hlbeg, hlbeglen,
                                         hlend, hlendlen) :
    UdmDSTRAppendRemoveHiLight(&ReturnValue->Value, str->str, str->length);
}


static void
UdmDocumentMethodPropertyInternal(UDM_PROG_EXECUTOR_STATE *state,
                                  UDM_DOCUMENT *Document,
                                  UDM_SECTION *Property,
                                  udm_bool_t html, udm_bool_t for_cached,
                                  const char *hlbeg, size_t hlbeglen,
                                  const char *hlend, size_t hlendlen)
{
  UDM_SECTION *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UDM_CONST_STR str;
  UDM_VALUE *Value;
  if (!(Value= UdmVarListFindValueByDSTR(&Document->Sections,
                                         &Property->Value)))
    return;
  Value->handler->GetConstStr(UdmValueDataPtr(Value), &str);
  UdmDocumentMethodPropertyCommon(state, ret,
                                  &str, html, for_cached,
                                  hlbeg, hlbeglen, hlend, hlendlen);
}


static void
UdmDocumentMethodProperty(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_DOCUMENT *Document= (UDM_DOCUMENT *) state->vreg[0].Data;
  UDM_SECTION *Property= (UDM_SECTION *) state->vreg[1].Data;
  UdmDocumentMethodPropertyInternal(state, Document, Property,
                                    UDM_FALSE, UDM_FALSE, "", 0, "", 0);
}


static void
UdmDocumentMethodPropertyHtml(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_DOCUMENT *Document= (UDM_DOCUMENT *) state->vreg[0].Data;
  UDM_SECTION *Property= (UDM_SECTION *) state->vreg[1].Data;
  UdmDocumentMethodPropertyInternal(state, Document, Property,
                                    UDM_TRUE, UDM_FALSE, "", 0, "", 0);
}


static void
UdmDocumentMethodContentHighlight(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UDM_DOCUMENT *Document= (UDM_DOCUMENT *) state->vreg[0].Data;
  UDM_SECTION *HlBeg= (UDM_SECTION *) state->vreg[1].Data;
  UDM_SECTION *HlEnd= (UDM_SECTION *) state->vreg[2].Data;

  UDM_CONST_STR str;
  UDM_VALUE *Value;
  if (!(Value= UdmVarListFindValue(&Document->Sections, "document")))
    return;
  Value->handler->GetConstStr(UdmValueDataPtr(Value), &str);
  UdmDocumentMethodPropertyCommon(state, ret, &str,
                                    UDM_FALSE, UDM_TRUE,
                                    HlBeg->Value.Val.str,
                                    HlBeg->Value.Val.length,
                                    HlEnd->Value.Val.str,
                                    HlEnd->Value.Val.length);
}


static UDM_VALUE_PROTOTYPE UdmDocumentMethodPrototypeDocumentStrings[]=
{
 {&UdmValueHandlerDocument},
 {&UdmValueHandlerSimple},
 {&UdmValueHandlerSimple},
 {&UdmValueHandlerSimple},
};


static void
UdmDocumentMethodCachedCopy(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_AGENT Agent;
  UDM_DOCUMENT *Document= (UDM_DOCUMENT*) state->vreg[0].Data;
  UDM_ENV *Env= ((UDM_ENV **) state->vreg[1].Data)[0];
  const UDM_SECTION *Section= (const UDM_SECTION*) state->vreg[2].Data;
  Env->errstr[0]= '\0';
  /*
    TODO34: if Env is reused for multiple queries, this UdmEnvPrepare()
    will load ispell multiple times. Fix this to avoid multiple loading.
  */
  UdmEnvPrepare(Env);
  UdmAgentInit(&Agent, Env, 0);
  
  UdmVarListAddQueryStringStrn(&Document->Sections, "query.",
                               UdmDSTRPtr(&Section->Value),
                               UdmDSTRLength(&Section->Value));
  UdmVarListAddQueryStringStrn(&Env->Vars, "query.",
                               UdmDSTRPtr(&Section->Value),
                               UdmDSTRLength(&Section->Value));
  UdmDocCachedCopyGet(&Agent, Document, &Document->Sections);
  UdmAgentFree(&Agent);
  UDM_FUNC_RETURN_INT(state, 0);
}


static void
UdmDocumentMethodDownload(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_AGENT Agent;
  UDM_DOCUMENT *Document= (UDM_DOCUMENT*) state->vreg[0].Data;
  UDM_ENV *Env= ((UDM_ENV **) state->vreg[1].Data)[0];
  const UDM_SECTION *Section= (const UDM_SECTION*) state->vreg[2].Data;
  const char *url= UdmDSTRPtr(&Section->Value);
  UdmAgentInit(&Agent, Env, 0);
  UDM_FUNC_RETURN_INT(state, UdmGetURLSimple(&Agent, Document, url));
  UdmAgentFree(&Agent);
}


static void
UdmDocumentMethodContent(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_DOCUMENT *Document= (UDM_DOCUMENT*) state->vreg[0].Data;
  UDM_CONST_STR str;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  if (UdmHTTPBufContentToConstStr(&Document->Buf, &str))
    UDM_BZERO(&str, sizeof(str));
  UdmDSTRAppend(&ReturnValue->Value, str.str, str.length);
}


static UDM_VALUE_PROTOTYPE UdmDocumentMethodPrototypeCachedCopy[]=
{
 {&UdmValueHandlerDocument},
 {&UdmValueHandlerEnv},
 {&UdmValueHandlerSimple}
};


static UDM_VALUE_PROTOTYPE UdmDocumentMethodPrototypeDownload[]=
{
 {&UdmValueHandlerDocument},
 {&UdmValueHandlerEnv},
 {&UdmValueHandlerSimple}
};


static const UDM_FUNCTION DocumentMethods[]=
{
  {{UDM_CSTR_WITH_LEN("cached")},
    {{&UdmValueHandlerInt}, 3, UdmDocumentMethodPrototypeCachedCopy},
    UdmDocumentMethodCachedCopy},
  {{UDM_CSTR_WITH_LEN("download")},
    {{&UdmValueHandlerInt}, 3, UdmDocumentMethodPrototypeDownload},
    UdmDocumentMethodDownload},
  {{UDM_CSTR_WITH_LEN("property")},
    {{&UdmValueHandlerSimple}, 2, UdmDocumentMethodPrototypeDocumentStrings},
    UdmDocumentMethodProperty},
  {{UDM_CSTR_WITH_LEN("content_highlight")},
    {{&UdmValueHandlerSimple}, 3, UdmDocumentMethodPrototypeDocumentStrings},
    UdmDocumentMethodContentHighlight},
  {{UDM_CSTR_WITH_LEN("content")},
    {{&UdmValueHandlerSimple}, 1, UdmDocumentMethodPrototypeDocumentStrings},
    UdmDocumentMethodContent},
  {{UDM_CSTR_WITH_LEN("property_html")},
    {{&UdmValueHandlerSimple}, 2, UdmDocumentMethodPrototypeDocumentStrings},
    UdmDocumentMethodPropertyHtml},
  {{0,0}, {{NULL}, 0, NULL}, NULL}
};


UDM_VALUE_HANDLER UdmValueHandlerDocument=
{
  UDM_VALUE_HANDLER_TYPE_DOCUMENT,
  UDM_VALUE_DATA_TYPE_DOCUMENT,
  UDM_VALUE_DATA_TYPE_STR,
  "DOCUMENT",
  UdmValueDocumentConstructor,
  UdmValueDocumentDestructor,
  UdmValueDocumentCopy,
  UdmValueDocumentVarSize,
  UdmValueDocumentValueSize,
  UdmValueDocumentDataOffset,
  UdmValueDocumentDataSize,
  UdmValueDocumentDataAlignment,
  UdmValueDocumentDump,
  UdmValueDocumentFlags,
  UdmValueDocumentSecno,
  UdmValueDocumentMemUsed,
  UdmValueDocumentGetConstStr,
  UdmValueDocumentGetBool,
  UdmValueDocumentGetInt,
  UdmValueDocumentGetUInt,
  UdmValueDocumentGetDouble,
  UdmValueDocumentSetInt,
  UdmValueDocumentSetDouble,
  UdmValueDocumentSetStrn,
  UdmValueDocumentPrintToFile,
  UdmValueDocumentPrintToDSTR,
  UdmValueDocumentConvertCharset,
  UdmValueDocumentSetConv,
  UdmValueDocumentAppendStrn,
  UdmValueDocumentAppendConv,
  DocumentMethods,
};
