/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/


#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_match.h"


typedef struct
{
  UDM_VALUE generic;
  UDM_EXCERPT_FRAGMENT Specific;
} UDM_VALUE_EXCERPT_FRAGMENT;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_EXCERPT_FRAGMENT Value;
} UDM_VAR_EXCERPT_FRAGMENT;


#define UDM_VALUE_EXCERPT_FRAGMENT_GET_SPECIFIC(x) \
  (&((UDM_VALUE_EXCERPT_FRAGMENT*) x)->Specific)
#define UDM_VALUE_EXCERPT_FRAGMENT_GET_SPECIFIC_CONST(x) \
  (&((const UDM_VALUE_EXCERPT_FRAGMENT*) x)->Specific)


static udm_rc_t
UdmValueExcerptFragmentDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueExcerptFragmentFlags(const char *Data)
{
  return UDM_VARFLAG_READONLY;
}


static udm_secno_t
UdmValueExcerptFragmentSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueExcerptFragmentMemUsed(const char *Data)
{
  return sizeof(UDM_VALUE_EXCERPT_FRAGMENT);
}


static void
UdmValueExcerptFragmentDestructor(char *Data)
{
  UDM_EXCERPT_FRAGMENT *Res= (UDM_EXCERPT_FRAGMENT *) Data;
  UdmExcerptFragmentFree(Res);
}


static udm_rc_t
UdmValueExcerptFragmentCopy(char *D, const char *S)
{
  UDM_ASSERT(0); /* TODO34: Add CloneVar(), with reference */
  memmove(D, S, sizeof(UDM_VALUE_EXCERPT_FRAGMENT));
  return UDM_OK;
}


static udm_rc_t
UdmValueExcerptFragmentConstructor(char *Data,
                                   const UDM_VALUE **args, size_t nargs)
{
  UDM_EXCERPT_FRAGMENT *Specific= (UDM_EXCERPT_FRAGMENT*) Data;
  UdmExcerptFragmentInit(Specific);
  return UDM_OK;
}


static size_t
UdmValueExcerptFragmentVarSize(void)
{
  return sizeof(UDM_VAR_EXCERPT_FRAGMENT);
}


static size_t
UdmValueExcerptFragmentValueSize(void)
{
  return sizeof(UDM_VALUE_EXCERPT_FRAGMENT);
}


static size_t
UdmValueExcerptFragmentDataOffset(void)
{
  return offsetof(UDM_VALUE_EXCERPT_FRAGMENT, Specific);
}


static size_t
UdmValueExcerptFragmentDataSize(void)
{
  return sizeof(UDM_EXCERPT_FRAGMENT);
}


static size_t
UdmValueExcerptFragmentDataAlignment(void)
{
  return sizeof(size_t);
}


static void
UdmValueExcerptFragmentGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  static const UDM_CONST_STR name= {UDM_CSTR_WITH_LEN("#ExcerptFragment#")};
  *Val= name;
}


static void
UdmValueExcerptFragmentGetBool(const char *Data, udm_bool_t *val)
{
  *val= UDM_TRUE;
}


static void
UdmValueExcerptFragmentGetInt(const char *Data, int *val)
{
  *val= 0;
}


static void
UdmValueExcerptFragmentGetUInt(const char *Data, unsigned int *val)
{
  *val= 0;
}


static void
UdmValueExcerptFragmentGetDouble(const char *Data, double *val)
{
  *val= 0.0;
}


static udm_rc_t
UdmValueExcerptFragmentSetInt(char *Data, int value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueExcerptFragmentSetDouble(char *Data, double value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueExcerptFragmentConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueExcerptFragmentAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueExcerptFragmentSetStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueExcerptFragmentSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                      const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueExcerptFragmentAppendConv(char *Data, size_t maxlen,
                                  UDM_CONV *cnv, int cnvflags,
                                  const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueExcerptFragmentPrintToFile(const char *Data, FILE *f)
{
  return UDM_OK;
}


static udm_rc_t
UdmValueExcerptFragmentPrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  return UDM_OK;
}


UDM_VALUE_HANDLER UdmValueHandlerExcerptFragment=
{
  UDM_VALUE_HANDLER_TYPE_EXCERPT_FRAGMENT,
  UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT,
  UDM_VALUE_DATA_TYPE_STR,
  "UdmResult",
  UdmValueExcerptFragmentConstructor,
  UdmValueExcerptFragmentDestructor,
  UdmValueExcerptFragmentCopy,
  UdmValueExcerptFragmentVarSize,
  UdmValueExcerptFragmentValueSize,
  UdmValueExcerptFragmentDataOffset,
  UdmValueExcerptFragmentDataSize,
  UdmValueExcerptFragmentDataAlignment,
  UdmValueExcerptFragmentDump,
  UdmValueExcerptFragmentFlags,
  UdmValueExcerptFragmentSecno,
  UdmValueExcerptFragmentMemUsed,
  UdmValueExcerptFragmentGetConstStr,
  UdmValueExcerptFragmentGetBool,
  UdmValueExcerptFragmentGetInt,
  UdmValueExcerptFragmentGetUInt,
  UdmValueExcerptFragmentGetDouble,
  UdmValueExcerptFragmentSetInt,
  UdmValueExcerptFragmentSetDouble,
  UdmValueExcerptFragmentSetStrn,
  UdmValueExcerptFragmentPrintToFile,
  UdmValueExcerptFragmentPrintToDSTR,
  UdmValueExcerptFragmentConvertCharset,
  UdmValueExcerptFragmentSetConv,
  UdmValueExcerptFragmentAppendStrn,
  UdmValueExcerptFragmentAppendConv,
  NULL,                    /* Methods */
};
